#include "StdAfx.h"
#include "Kinematics.h"

CKinematics::CKinematics(int taskSize) :
	_desired(taskSize), _current(taskSize)
{
	_desired.zero ();
	_current.zero ();
}

CKinematics::~CKinematics(void)
{
}

void CKinematics::AttachJoint (ObjectBase *obj)
{
	_objectList.push_back (obj);
}

dVector CKinematics::GetJointAngle ()
{
	dVector ang(_objectList.size());

	for (unsigned int i=0; i<_objectList.size(); ++i) {
		ObjectJoint *joint = dynamic_cast<ObjectJoint *>(_objectList[i]);
		ang[i] = (joint) ? joint->_q : 0.;
	}
	return ang;
}

void CKinematics::ClampMag(dVector &w, double maxMag)
{
	double mag = Norm2 (w);
	
	if (maxMag < mag){
		w *= maxMag/mag;
	}
}

dMatrix CKinematics::PseudoInverseWithDLS (dMatrix &J, double dlsLambda)
{
	// Moore-Penrose pseudo-inverse:
	// case m < n: invA = ~A * !(A * ~A)
	if (J.rowno() <= J.colno()) {
		dMatrix I(J.rowno(), J.rowno()); 
		I.unit();
		// Damped least-square method
		return ~J * !( J*~J + dlsLambda*dlsLambda*I );
	}
	// case m > n: invA = !(~A * A) * ~A
	else {
		dMatrix I(J.colno(), J.colno()); 
		I.unit();
		// Damped least-square method
		return !( ~J*J + dlsLambda*dlsLambda*I ) * ~J;
	}
}


double CKinematics::WeightFunction (double lo, double hi, double q)
{
 	if (lo < hi) {
		if (q < lo) q = lo;
		if (q > hi) q = hi;

		// w = 2(q - lo)/(hi - lo) -1		// ȸ  [lo, hi] [-1, 1]  ȯ q ġ 
		// w = w^32	// U   ٲ۴.
		double w = 2*(q - lo)/(hi - lo) - 1;
		w *= w; w *= w; w *= w; w *= w; w *= w; 
		return w;
	}
	else {
		return 1.;
	}
}

dMatrix CKinematics::WeightMatrix (unsigned int dof)
{
	dMatrix W(dof, dof);
	W.unit();

/*
	if (_weightCur.size() != dof) {
		_weightCur.resize (dof);
		
		for (unsigned int i=0; i<dof; ++i) {
			JointNullMotion *joint = _jointList[i];
			_weightCur[i] = joint->w;
		}
	}

	_weightPrv = _weightCur;

	for (unsigned int i=0; i<_jointList.size () && i<dof; ++i) {
		JointNullMotion *joint = _jointList[i];

		// Joint ȸ(q) 谪(lo:limits low, hi:limits high) ó   weight Ѵ
		_weightCur[i] = joint->w*(1. - WeightFunction(joint->lo, joint->hi, _kin->GetJointAngle(i)));
		// 1 step  weight(_weightPrv)   weight(_weightCur) Ͽ
		// weight    ó  ̹Ƿ weight  ״ ΰ
		// weight   迡  ̹Ƿ weight  ְ Ѵ.
		W(i,i) = (_weightPrv[i] < _weightCur[i]) ? joint->w : _weightCur[i];
	}
*/

	for (unsigned int i=0; i<_objectList.size(); ++i) {
		ObjectJoint *joint = dynamic_cast<ObjectJoint *>(_objectList[i]);
		W(i,i) = (joint) ? joint->_weight : 0.;
	}
	return W;
}

dVector CKinematics::NullSpace ()
{
	dVector h(_objectList.size ());

	for (unsigned int i=0; i<_objectList.size (); ++i) {
		ObjectJoint *joint = dynamic_cast<ObjectJoint *>(_objectList[i]);

		// Joint ȸ(q) (basis: ڰ  κ ⺻ ڼ) ó ϵ Ѵ
		h[i] = (joint) ? joint->_median - joint->_q : 0.;
	}

	return h;
}

dVector CKinematics::SolveJTR (double minAlpha, double clampTask)
{
	Forward ();
	dMatrix J = Jacobian ();
	dVector de = Error ();

	ClampMag (de, clampTask);

	// Compute alpha constant as quotient of 2 inner products as suggested by
	// Buss & Kim, to minimze de after caller updates theta
	dVector JJtde = (J * ~J) * de;
	double alpha = Dot(de, JJtde)/Dot (JJtde, JJtde);

	if (alpha < minAlpha) {
		alpha = minAlpha;
	}

	return alpha * ~J * de;
}

dVector CKinematics::SolveDLS (double dlsLambda, double clampTask, double clampNull)
{
	Forward ();
	dMatrix J = Jacobian ();
	dVector de = Error ();

	ClampMag (de, clampTask);

	dMatrix invJ = PseudoInverseWithDLS (J, dlsLambda);

	dVector dq = invJ * de;

	if (0. < clampNull) {
		dVector null = NullSpace ();

		ClampMag (null, clampNull);

		dMatrix I(J.colno(), J.colno());
		I.unit();

		dq += (I - invJ*J)*null;
	}

	return dq;
}

void CKinematics::Apply (dVector &dq)
{
	for (unsigned int i=0; i<_objectList.size () && i<dq.size(); ++i) {
		ObjectJoint *joint = dynamic_cast<ObjectJoint *>(_objectList[i]);
		joint->_q += dq[i];
		if (joint && joint->_limit_lo < joint->_limit_hi) {
			joint->_q = Limit (joint->_q, joint->_limit_lo, joint->_limit_hi);
		}
	}
}

