#include "StdAfx.h"
#include "MultiTasks.h"

CMultiTasks::CMultiTasks(void)
: CKinematics(0)
{
}

CMultiTasks::~CMultiTasks(void)
{
}

dVector CMultiTasks::AttachVector (dVector &a, dVector &b)
{
	dVector c(a.size() + b.size());
	
	for (unsigned int i=0; i<a.size(); ++i) {
		c[i] = a[i];
	}
	for (unsigned int i=0; i<b.size(); ++i) {
		c[i + a.size()] = b[i];
	}
	return c;
}

int CMultiTasks::FindIndex (vector<ObjectBase *> &objectList, const string &name) 
{
	for (unsigned int i=0; i<objectList.size(); ++i) {
		if (objectList[i]->_name == name) {
			return i;
		}
	}
	return -1;
}

dMatrix CMultiTasks::ReformMatrixByIndex (dMatrix &m, int colno, vector<unsigned int> &indexList)
{
	dMatrix n(m.rowno (), colno);
	n.zero ();

	for (unsigned int r=0; r<m.rowno(); ++r) {
		for (unsigned int c=0; c<m.colno(); ++c) {
			n(r, indexList[c]) = m(r, c);
		}
	}
	return n;
}

void CMultiTasks::AddTask (CKinematics *task)
{
	vector<unsigned int> indexList;
	vector<ObjectBase *> &objectList = task->GetObjectList ();

	for (unsigned int i=0; i<objectList.size(); ++i) {
		ObjectBase *obj = objectList[i];
		int index = FindIndex (_objectList, obj->_name);
		if (index == -1) {
			AttachJoint (obj);
			index = _objectList.size() - 1;
		}
		indexList.push_back (index);
	}

	_taskList.push_back (make_pair (task, indexList));
}

dMatrix CMultiTasks::Forward ()
{
	for (unsigned int i=0; i<_taskList.size(); ++i) {
		CKinematics *kin = _taskList[i].first;

		kin->Forward ();
	}

	dVector c(0), d(0);
	for (unsigned int i=0; i<_taskList.size(); ++i) {
		CKinematics *kin = _taskList[i].first;

		c = AttachVector (c, kin->GetCurrent ());
		d = AttachVector (d, kin->GetDesired ());
	}
	_current = c;
	_desired = d;

	return dMatrix (0, 0);
}

dVector CMultiTasks::Error ()
{
	dVector error(0);

	for (unsigned int i=0; i<_taskList.size(); ++i) {
		CKinematics *kin = _taskList[i].first;

		error = AttachVector (error, kin->Error ());
	}
	return error;
}

dMatrix CMultiTasks::Jacobian ()
{
	dMatrix Js(_current.size(), _objectList.size());
	Js.zero ();

	int rr = 0, cc = 0;
	for (unsigned int i=0; i<_taskList.size(); ++i) {
		CKinematics *kin = _taskList[i].first;
		vector<unsigned int> &indexList = _taskList[i].second;

		dMatrix J = kin->Jacobian ();

		for (unsigned int r=0; r<J.rowno(); ++r) {
			for (unsigned int c=0; c<J.colno(); ++c) {
				cc = indexList[c];
				Js(rr, cc) = J(r, c);
			}
			++rr;
		}
	}
	return Js;
}

dVector CMultiTasks::SolveChi (double dlsLambda, double clampTask)
{
	int dof = _objectList.size();
	dVector qd(dof);
	qd.zero ();

	for (unsigned int i=0; i<_taskList.size (); ++i) {
		CKinematics *kin = _taskList[i].first;
		vector<unsigned int> &indexList = _taskList[i].second;

		kin->Forward ();
		dMatrix J = kin->Jacobian ();
		dVector de = kin->Error ();

		ClampMag (de, clampTask);
	
		J = ReformMatrixByIndex (J, dof, indexList);
		dMatrix invJ = PseudoInverseWithDLS (J, dlsLambda);

		dMatrix I(J.colno(),J.colno()); I.unit();
		qd = invJ*de + (I - invJ*J)*qd;
	}
	return qd;
}

dVector CMultiTasks::SolveNak (double dlsLambda, double clampTask)
{
	int dof = _objectList.size();
	dVector qd (dof);		// qd = 0
	qd.zero ();

	dMatrix P(dof, dof);	// P = I
	P.unit();

	for (int i=_taskList.size ()-1; 0<=i; --i) {
		CKinematics *kin = _taskList[i].first;
		vector<unsigned int> &indexList = _taskList[i].second;

		kin->Forward ();
		dMatrix J = kin->Jacobian ();
		dVector de = kin->Error ();

		ClampMag (de, clampTask);

		J = ReformMatrixByIndex (J, dof, indexList);

		// qd  = qd_prv + !J^ * (dr - J * qd_prv)
		// P   = P_prv - !J^ * J^
		// J^  = J * P_prv
		dMatrix Jhat = J*P;
		dMatrix I(J.rowno (), J.rowno ()); I.unit();
		dMatrix invJhat = PseudoInverseWithDLS (Jhat, dlsLambda);
		P = P - invJhat*Jhat;
		qd = qd + invJhat*(de - J*qd);
	}
	return qd;
}
